// This is free software without any guarantee.
#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <time.h>

HHOOK hKeyHook, hMouseHook;
FILE *f;
int wpos = 0, v[10];
struct tm period;
time_t periodtime; // := mktime (&period)

struct tm *ThisPeriod (time_t t)
{
  static struct tm tp;
  do { // Find the start of the current time period.
    memcpy (&tp, localtime (&t), sizeof (tp));
    t -= 7200;
  } while (tp.tm_hour < 2);
  tp.tm_min = tp.tm_sec = tp.tm_wday = tp.tm_yday = tp.tm_isdst = 0;
  tp.tm_hour = 2; // Set every unused field to 0 to enable memcmp()
  return &tp;
}

void Write (void)
{
  fseek (f, wpos, SEEK_SET);
  fprintf (f, "%02d/%02d/%02d %02d:%02d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d\r\n",
    period.tm_mon + 1, period.tm_mday, period.tm_year % 100, period.tm_hour,
    period.tm_min, v[0], v[1], v[2], v[3], v[4], v[5], v[6], v[7], v[8],v[9]);
}

typedef enum { alnumKey, otherKey, mouseClick, mouseMove } eventType;

void Log (eventType event)
{
  static int wtime = 0, mtime = 0, t4 = 0, t5 = 0, t6 = 0, t7 = 0;
  time_t nout = time (NULL);
  struct tm *nou = ThisPeriod (nout);
  if (!f) return;
  if (memcmp (nou, &period, sizeof (period)) != 0) {
    Write ();
    wpos = ftell (f);
    
    memset (v, 0, sizeof (v));
    wtime = 0;
    memcpy (&period, nou, sizeof (period));
    periodtime = mktime (nou);
    v[9] = nout - periodtime;
  }
  if (event == alnumKey || event == otherKey) {
    v[0]++;
    if (t4 + 40 >= nout) v[4] += nout - t4;
    if (t5 + 4 >= nout) v[5] += nout - t5;
    t4 = t5 = nout;
    if (event == alnumKey) {
      v[1]++;
      if (t6 + 4 >= nout) v[6] += nout - t6;
      t6 = nout;
    }
  }
  if (event == mouseClick) v[2]++;
  if (event == mouseClick || event == mouseMove) {
    if (mtime + 4 >= nout) v[3] += nout - mtime;
    mtime = nout;
  }
  if (t7 + 4 >= nout) v[7] += nout - t7;
  t7 = nout;
  v[8] = nout - periodtime - v[9];
  
  if (wtime < nout) {
    int tail = 0;
    Write ();
    fseek (f, 0, SEEK_CUR); // Must we still seek() between write() and read ?
    while (getc (f) != EOF) tail++;
    for (fseek (f, -tail, SEEK_CUR); tail > 0; tail -= 2) fprintf (f, "\r\n");
    fflush (f);
    wtime = nout + 30; // sync() every 30 seconds or so.
  }
}

LRESULT CALLBACK KeyHookMsg (int nCode, WPARAM w, LPARAM l)
{
  KBDLLHOOKSTRUCT *ev = (KBDLLHOOKSTRUCT*) l;
  if (nCode == HC_ACTION && (w == WM_SYSKEYDOWN || w == WM_KEYDOWN)) {
  // Don't count PeekMsg calls. Don't count release of the key.
    Log (isalnum (ev->vkCode) ? alnumKey : otherKey);
  }
  
  return CallNextHookEx (hKeyHook, nCode, w, l);
  // My (old) help file does not say when NOT to call CallNextHookEx, so I'm
  // calling it every time.
}

LRESULT CALLBACK Mouse (int nCode, WPARAM w, LPARAM l)
{
  MOUSEHOOKSTRUCT *ev = (MOUSEHOOKSTRUCT*) l;
  if (nCode == HC_ACTION) {
    Log (w == WM_LBUTTONDOWN || w == WM_LBUTTONDBLCLK || w == WM_RBUTTONDOWN
      || w == WM_RBUTTONDBLCLK || w == WM_MBUTTONDOWN || w == WM_MBUTTONDBLCLK
      ? mouseClick : mouseMove);
  }
  return CallNextHookEx (hMouseHook, nCode, w, l);
}

NOTIFYICONDATA nid;

LRESULT CALLBACK WindowProc( HWND hwnd, UINT umsg, WPARAM w, LPARAM l)
{ // This window is always hidden.
  if (umsg == WM_DESTROY) PostQuitMessage (0);
  else if (umsg == nid.uCallbackMessage) Shell_NotifyIcon (NIM_MODIFY, &nid);
  else return DefWindowProc( hwnd, umsg, w, l);
  return FALSE;
}

int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
    LPSTR lpCmdLine, int nCmdShow)
{
  MSG msg;
  HWND tray;
  WNDCLASS        wc;
  struct tm tmptm; // This is used to prevent mktime for changing 'period'

  memset (&wc, 0, sizeof (wc));
  wc.style = CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc = WindowProc;
  wc.hInstance = hInstance;
  wc.hIcon = LoadIcon (hInstance, MAKEINTRESOURCE(1));
  wc.hCursor = LoadCursor( NULL, IDC_ARROW );
  wc.hbrBackground = NULL;
  wc.lpszClassName = "Chronologer";
  wc.lpszMenuName = "Chronologer";
  if (!RegisterClass (&wc)) return FALSE;
  
  tray = CreateWindow (wc.lpszClassName, wc.lpszClassName,
    WS_OVERLAPPEDWINDOW, 0, 0, 0, 0, NULL, NULL, hInstance, NULL);
  ShowWindow (tray, SW_HIDE);
  UpdateWindow (tray);
  memset (&nid, 0, sizeof (nid));
  nid.cbSize = sizeof (nid);
  nid.hIcon = wc.hIcon;
  nid.hWnd = tray;
  nid.uCallbackMessage = WM_USER + 3;
  nid.uID = 0;
  nid.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
  strcpy (nid.szTip, "Cronologer - Measuring PC usage");
  Shell_NotifyIcon (NIM_ADD, &nid);
 
  hKeyHook = SetWindowsHookEx (WH_KEYBOARD_LL, (HOOKPROC) KeyHookMsg,
    hInstance, 0); // What is the _LL stuff ?
  hMouseHook = SetWindowsHookEx (WH_MOUSE_LL, (HOOKPROC) Mouse, hInstance, 0);
  if ((f = fopen ("chronologer.csv", "r+b"))) {
    int ch1, ch2;
    for (wpos = -2; ; wpos--) { // Find the beginning of fist entry
      if (fseek (f, wpos, SEEK_END) != 0 || (ch1 = getc (f)) == EOF ||
      (ch2 = getc (f)) == EOF) {
        wpos = 0; // Only 1 entry or less
        break;
      }
      if (ch1 == '\n' && ch2 != '\n' && ch2 != '\r') {
        wpos = ftell (f) - 1;
        break;
      }
    }
    fseek (f, wpos, SEEK_SET);
    if (fscanf (f, "%d/%d/%d %d:%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d",
      &period.tm_mon, &period.tm_mday, &period.tm_year, &period.tm_hour,
      &period.tm_min,
      v + 0, v + 1, v + 2, v + 3, v + 4, v + 5, v + 6, v + 7, v + 8, v + 9
      ) != 15) memcpy (&period, ThisPeriod (time (NULL)), sizeof (period));
    else { // No error occured.
      period.tm_mon--; // Fill in every field to allow memcmp ()
      period.tm_year = (period.tm_year + 30) % 100 + 70;
      period.tm_sec = period.tm_wday = period.tm_yday = period.tm_isdst = 0;
    }
  }
  else if ((f = fopen ("chronologer.csv", "wb")) != NULL) {
    memcpy (&period, ThisPeriod (time (NULL)), sizeof (period));
  }
  else MessageBox (NULL, lpCmdLine, "Unable to open / create .CSV file", 0);
  periodtime = mktime (memcpy (&tmptm, &period, sizeof (tmptm)));
 
  while (GetMessage (&msg, NULL, 0, 0)) {
    TranslateMessage (&msg);
    DispatchMessage (&msg);
  }
  return 0;
}
